import unittest

import pyvex
from pyvex import IRSB, ffi, lift
from pyvex.errors import PyVEXError
from pyvex.lifting.util import GymratLifter, Instruction, JumpKind


# pylint: disable=R0201
# pylint: disable=C0115
class TestLift(unittest.TestCase):
    def test_partial_lift(self):
        """This tests that gymrat correctly handles the case where an
        instruction is longer than the remaining input.
        """

        class NOP(Instruction):
            name = "nop"
            bin_format = "0000111100001111"

            def compute_result(self, *args):
                pass

        class NOPLifter(GymratLifter):
            instrs = [NOP]

        lifter = NOPLifter(pyvex.ARCH_AMD64, 0)
        # this should not throw an exception
        block = lifter.lift("\x0F\x0Fa")
        assert block.size == 2
        assert block.instructions == 1
        assert block.jumpkind == JumpKind.NoDecode

    def test_skipstmts_toomanyexits(self):
        # https://github.com/angr/pyvex/issues/153

        old_exit_limit = IRSB.MAX_EXITS
        IRSB.MAX_EXITS = 32

        bytes_ = bytes.fromhex(
            "0DF1B00B2EAB94E8030008938BE803000DF1C0089AE8030083E"
            "80300019B0DF1F00A339AE669E26193E8030085E8030098E803"
            "0083E80300069B95E8030088E80300A26993E803004A9200236"
            "3622362A361E362A36238AC029A069484E8030012AC09982993"
            "28932B9303C885E8030092E8030084E803009AE8030082E8030"
            "02A460A9D26993E910B9941910D9942910C992A93409548AD43"
            "9194E803008AE8030027983F9927913F909BE803000DF5887B2"
            "69335938BE803000DF58C7B089903C98BE8030098E8030084E8"
            "030095E8030088E803004B993391329394E8030034933793369"
            "3069C059B4C93049B4E9350ABCDF834C1CDF83CE185E8030094"
            "E803004B9683E8030015A94498C4F7E2EA "
        )
        arch = pyvex.ARCH_ARM_LE
        # Lifting the first four bytes will not cause any problem. Statements should be skipped as expected
        b = IRSB(bytes_[:34], 0xC6951, arch, opt_level=1, bytes_offset=5, skip_stmts=True)
        assert len(b.exit_statements) > 0
        assert not b.has_statements

        # Lifting the entire block will cause the number of exit statements go
        # beyond the limit (currently 32). PyVEX will
        # automatically relift this block without skipping the statements
        b = IRSB(bytes_, 0xC6951, arch, opt_level=1, bytes_offset=5, skip_stmts=True)
        assert b.statements is not None
        assert len(b.exit_statements) > 32

        # Restore the setting
        IRSB.MAX_EXITS = old_exit_limit

    def test_max_bytes(self):
        data = bytes.fromhex("909090909090c3")
        arch = pyvex.ARCH_X86
        assert lift(data, 0x1000, arch, max_bytes=None).size == len(data)
        assert lift(data, 0x1000, arch, max_bytes=len(data) - 1).size == len(data) - 1
        assert lift(data, 0x1000, arch, max_bytes=len(data) + 1).size == len(data)

        data2 = ffi.from_buffer(data)
        self.assertRaises(PyVEXError, lift, data2, 0x1000, arch)
        assert lift(data2, 0x1000, arch, max_bytes=len(data)).size == len(data)
        assert lift(data2, 0x1000, arch, max_bytes=len(data) - 1).size == len(data) - 1


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
import pyvex


def test_s390x_vl():
    arch = pyvex.ARCH_S390X
    irsb = pyvex.lift(b"\xe7\x40\x90\xa8\x00\x06", 0x11C6C9E, arch)  # vl %v4, 0xa8(%r9)
    irsb_str = str(irsb)

    assert "GET:I64(r9)" in irsb_str
    assert "Add64(0x00000000000000a8" in irsb_str
    assert "LDbe:V128" in irsb_str
    assert "PUT(v4) =" in irsb_str
    assert irsb.jumpkind == "Ijk_Boring"


if __name__ == "__main__":
    test_s390x_vl()



class PyVEXError(Exception):
    pass


class SkipStatementsError(PyVEXError):
    pass


#
# Exceptions and notifications that post-processors can raise
#


class LiftingException(Exception):
    pass


class NeedStatementsNotification(LiftingException):
    """
    A post-processor may raise a NeedStatementsNotification if it needs to work with statements, but the current IRSB
    is generated without any statement available (skip_stmts=True). The lifter will re-lift the current block with
    skip_stmts=False upon catching a NeedStatementsNotification, and re-run the post-processors.

    It's worth noting that if a post-processor always raises this notification for every basic block without statements,
    it will essentially disable the skipping statement optimization, and it is bad for performance (especially for
    CFGFast, which heavily relies on this optimization). Post-processor authors are encouraged to at least filter the
    IRSBs based on available properties (jumpkind, next, etc.). If a post-processor must work with statements for the
    majority of IRSBs, the author should implement it in PyVEX in C for the sake of a better performance.
    """

    pass



"""
PyVEX provides an interface that translates binary code into the VEX intermediate representation (IR).
For an introduction to VEX, take a look here: https://docs.angr.io/advanced-topics/ir
"""

__version__ = "9.2.118"

from . import const, expr, stmt
from .arches import (
    ARCH_AMD64,
    ARCH_ARM64_BE,
    ARCH_ARM64_LE,
    ARCH_ARM_BE,
    ARCH_ARM_BE_LE,
    ARCH_ARM_LE,
    ARCH_MIPS32_BE,
    ARCH_MIPS32_LE,
    ARCH_MIPS64_BE,
    ARCH_MIPS64_LE,
    ARCH_PPC32,
    ARCH_PPC64_BE,
    ARCH_PPC64_LE,
    ARCH_RISCV64_LE,
    ARCH_S390X,
    ARCH_X86,
)
from .block import IRSB, IRTypeEnv
from .const import get_type_size, get_type_spec_size, tag_to_const_class
from .enums import (
    IRCallee,
    IRRegArray,
    VEXObject,
    default_vex_archinfo,
    get_enum_from_int,
    get_int_from_enum,
    irop_enums_to_ints,
    vex_endness_from_string,
)
from .errors import PyVEXError
from .expr import get_op_retty
from .lifting import lift, lifters
from .native import ffi, pvc

# aliases....
IRStmt = stmt
IRExpr = expr
IRConst = const


__all__ = [
    "const",
    "expr",
    "stmt",
    "IRSB",
    "IRTypeEnv",
    "get_type_size",
    "get_type_spec_size",
    "irop_enums_to_ints",
    "tag_to_const_class",
    "IRCallee",
    "IRRegArray",
    "VEXObject",
    "default_vex_archinfo",
    "get_enum_from_int",
    "get_int_from_enum",
    "vex_endness_from_string",
    "PyVEXError",
    "get_op_retty",
    "lift",
    "lifters",
    "ffi",
    "pvc",
    "IRStmt",
    "IRExpr",
    "IRConst",
    "ARCH_X86",
    "ARCH_AMD64",
    "ARCH_ARM_BE",
    "ARCH_ARM_BE_LE",
    "ARCH_ARM_LE",
    "ARCH_ARM64_LE",
    "ARCH_ARM64_BE",
    "ARCH_PPC32",
    "ARCH_PPC64_BE",
    "ARCH_PPC64_LE",
    "ARCH_S390X",
    "ARCH_MIPS32_BE",
    "ARCH_MIPS32_LE",
    "ARCH_MIPS64_BE",
    "ARCH_MIPS64_LE",
    "ARCH_RISCV64_LE",
]



import re

from .enums import VEXObject, get_enum_from_int
from .errors import PyVEXError
from .native import ffi, pvc


# IRConst hierarchy
class IRConst(VEXObject):
    __slots__ = ["_value"]

    type: str
    size: int | None = None
    tag: str
    c_constructor = None
    _value: int

    def pp(self):
        print(self.__str__())

    @property
    def value(self) -> int:
        return self._value

    @staticmethod
    def _from_c(c_const):
        if c_const[0] == ffi.NULL:
            return None

        tag = get_enum_from_int(c_const.tag)

        try:
            return tag_to_const_class(tag)._from_c(c_const)
        except KeyError:
            raise PyVEXError("Unknown/unsupported IRConstTag %s\n" % tag)

    _translate = _from_c

    @classmethod
    def _to_c(cls, const):
        # libvex throws an exception when constructing a U1 with a value other than 0 or 1
        if const.tag == "Ico_U1" and const.value not in (0, 1):
            raise PyVEXError("Invalid U1 value: %d" % const.value)

        try:
            return cls.c_constructor(const.value)
        except KeyError:
            raise PyVEXError("Unknown/unsupported IRConstTag %s]n" % const.tag)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self._value == other._value

    def __hash__(self):
        return hash((type(self), self._value))


class U1(IRConst):
    __slots__: list[str] = []

    type = "Ity_I1"
    size = 1
    tag = "Ico_U1"
    op_format = "1"
    c_constructor = pvc.IRConst_U1

    def __init__(self, value):
        self._value = value

    def __str__(self):
        return "%d" % self.value

    @staticmethod
    def _from_c(c_const):
        return U1(c_const.Ico.U1)


class U8(IRConst):
    __slots__: list[str] = []

    type = "Ity_I8"
    size = 8
    tag = "Ico_U8"
    op_format = "8"
    c_constructor = pvc.IRConst_U8

    def __init__(self, value):
        self._value = value

    def __str__(self):
        return "0x%02x" % self.value

    @staticmethod
    def _from_c(c_const):
        return _U8_POOL[c_const.Ico.U8]


_U8_POOL = [U8(i) for i in range(256)]


class U16(IRConst):
    __slots__: list[str] = []

    type = "Ity_I16"
    size = 16
    tag = "Ico_U16"
    op_format = "16"
    c_constructor = pvc.IRConst_U16

    def __init__(self, value):
        self._value = value

    def __str__(self):
        return "0x%04x" % self.value

    @staticmethod
    def _from_c(c_const):
        val = c_const.Ico.U16
        if val < 1024:
            return _U16_POOL[val]
        if val >= 0xFC00:
            return _U16_POOL[val - 0xFC00 + 1024]
        return U16(val)


_U16_POOL = [U16(i) for i in range(1024)] + [U16(i) for i in range(0xFC00, 0xFFFF + 1)]


class U32(IRConst):
    __slots__: list[str] = []

    type = "Ity_I32"
    size = 32
    tag = "Ico_U32"
    op_format = "32"
    c_constructor = pvc.IRConst_U32

    def __init__(self, value: int):
        self._value = value

    def __str__(self):
        return "0x%08x" % self.value

    @staticmethod
    def _from_c(c_const):
        val = c_const.Ico.U32
        if val < 1024:
            return _U32_POOL[val]
        if val >= 0xFFFFFC00:
            return _U32_POOL[val - 0xFFFFFC00 + 1024]
        return U32(val)


_U32_POOL = [U32(i) for i in range(1024)] + [U32(i) for i in range(0xFFFFFC00, 0xFFFFFFFF + 1)]


class U64(IRConst):
    __slots__: list[str] = []

    type = "Ity_I64"
    size = 64
    tag = "Ico_U64"
    op_format = "64"
    c_constructor = pvc.IRConst_U64

    def __init__(self, value):
        self._value = value

    def __str__(self):
        return "0x%016x" % self.value

    @staticmethod
    def _from_c(c_const):
        val = c_const.Ico.U64
        if val < 1024:
            return _U64_POOL[val]
        if val >= 0xFFFFFFFFFFFFFC00:
            return _U64_POOL[val - 0xFFFFFFFFFFFFFC00 + 1024]
        return U64(val)


_U64_POOL = [U64(i) for i in range(1024)] + [U64(i) for i in range(0xFFFFFFFFFFFFFC00, 0xFFFFFFFFFFFFFFFF + 1)]

# Integer Type Imagination
class_cache = {1: U1, 8: U8, 16: U16, 32: U32, 64: U64}


def vex_int_class(size):
    try:
        return class_cache[size]
    except KeyError:

        class VexInt(IRConst):
            type = "Ity_I%d" % size
            tag = "Ico_U%d" % size
            op_format = str(size)

            def __init__(self, value):
                IRConst.__init__(self)
                self._value = value

            def __str__(self):
                return f"(0x{self.value:x} :: {self.type})"

        VexInt.__name__ = "U%d" % size
        class_cache[size] = VexInt
        return VexInt


class F32(IRConst):
    __slots__: list[str] = []

    type = "Ity_F32"
    tag = "Ico_F32"
    op_format = "F32"
    c_constructor = pvc.IRConst_F32

    def __init__(self, value):
        self._value = value

    def __str__(self):
        return "%f" % self.value

    @staticmethod
    def _from_c(c_const):
        return F32(c_const.Ico.F32)


class F32i(IRConst):
    __slots__: list[str] = []

    type = "Ity_F32"
    tag = "Ico_F32i"
    op_format = "F32"
    c_constructor = pvc.IRConst_F32i

    def __init__(self, value):
        self._value = value

    def __str__(self):
        return "%f" % self.value

    @staticmethod
    def _from_c(c_const):
        return F32i(c_const.Ico.F32)


class F64(IRConst):
    __slots__: list[str] = []

    type = "Ity_F64"
    tag = "Ico_F64"
    op_format = "F64"
    c_constructor = pvc.IRConst_F64

    def __init__(self, value):
        self._value = value

    def __str__(self):
        return "%f" % self.value

    @staticmethod
    def _from_c(c_const):
        return F64(c_const.Ico.F64)


class F64i(IRConst):
    __slots__: list[str] = []

    type = "Ity_F64"
    tag = "Ico_F64i"
    op_format = "F64"
    c_constructor = pvc.IRConst_F64i

    def __init__(self, value):
        self._value = value

    def __str__(self):
        return "%f" % self.value

    @staticmethod
    def _from_c(c_const):
        return F64i(c_const.Ico.F64)


class V128(IRConst):
    __slots__: list[str] = []

    type = "Ity_V128"
    tag = "Ico_V128"
    op_format = "V128"
    c_constructor = pvc.IRConst_V128

    def __init__(self, value):
        self._value = value

    def __str__(self):
        return "%x" % self.value

    # vex doesn't store a full 128 bit constant, instead it stores 1 bit per 8 bits of data
    # and duplicates each bit 8 times
    @staticmethod
    def _from_c(c_const):
        base_const = c_const.Ico.V128
        real_const = 0
        for i in range(16):
            if (base_const >> i) & 1 == 1:
                real_const |= 0xFF << (8 * i)
        return V128(real_const)


class V256(IRConst):
    __slots__: list[str] = []

    type = "Ity_V256"
    tag = "Ico_V256"
    op_format = "V256"
    c_constructor = pvc.IRConst_V256

    def __init__(self, value):
        self._value = value

    def __str__(self):
        return "%x" % self.value

    # see above
    @staticmethod
    def _from_c(c_const):
        base_const = c_const.Ico.V256
        real_const = 0
        for i in range(32):
            if (base_const >> i) & 1 == 1:
                real_const |= 0xFF << (8 * i)
        return V256(real_const)


predefined_types = [U1, U8, U16, U32, U64, F32, F32i, F64, F64i, V128, V256]
predefined_types_map = {c.type: c for c in predefined_types}
predefined_classes_map = {c.tag: c for c in predefined_types}

# precompiled regexes
int_ty_re = re.compile(r"Ity_I\d+")
int_tag_re = re.compile(r"Ico_U\d+")
tag_size_re = re.compile(r"Ico_[UFV](?P<size>\d+)i?")


def is_int_ty(ty):
    m = int_ty_re.match(ty)
    return m is not None


def is_int_tag(tag):
    m = int_tag_re.match(tag)
    return m is not None


def get_tag_size(tag):
    m = tag_size_re.match(tag)
    if m is None:
        raise ValueError("Tag %s does not have size" % tag)
    return int(m.group("size"))


type_str_re = re.compile(r"Ity_[IFDV](?P<size>\d+)")
type_tag_str_re = re.compile(r"[IFDV]?(?P<size>\d+)[SU]?")


def get_type_size(ty):
    """
    Returns the size, in BITS, of a VEX type specifier
    e.g., Ity_I16 -> 16

    :param ty:
    :return:
    """
    m = type_str_re.match(ty)
    if m is None:
        raise ValueError("Type %s does not have size" % ty)
    return int(m.group("size"))


def get_type_spec_size(ty):
    """
    Get the width of a "type specifier"
    like I16U
    or F16
    or just 16
    (Yes, this really just takes the int out.  If we must special-case, do it here.
    :param tyspec:
    :return:
    """
    m = type_tag_str_re.match(ty)
    if m is None:
        raise ValueError("Type specifier %s does not have size" % ty)
    return int(m.group("size"))


def ty_to_const_class(ty):
    try:
        return predefined_types_map[ty]
    except KeyError:
        if is_int_ty(ty):
            size = get_type_size(ty)
            return vex_int_class(size)
        else:
            raise ValueError("Type %s does not exist" % ty)


def tag_to_const_class(tag):
    try:
        return predefined_classes_map[tag]
    except KeyError:
        if is_int_tag(tag):
            size = get_tag_size(tag)
            return vex_int_class(size)
        else:
            raise ValueError("Tag %s does not exist" % tag)



import pyvex


def test_s390x_lochi():
    arch = pyvex.ARCH_S390X
    irsb = pyvex.lift(b"\xec\x18\xab\xcd\x00\x42", 0x400400, arch)  # lochi %r1,0xabcd,8
    irsb_str = str(irsb)

    assert "s390_calculate_cond(0x0000000000000008" in irsb_str
    assert "PUT(r1_32) = 0xffffabcd" in irsb_str
    assert irsb.jumpkind in "Ijk_Boring"


if __name__ == "__main__":
    test_s390x_lochi()



from .gym import AARCH64Spotter, AMD64Spotter, ARMSpotter, X86Spotter
from .libvex import LIBVEX_SUPPORTED_ARCHES, LibVEXLifter
from .lift_function import lift, lifters, register
from .lifter import Lifter
from .post_processor import Postprocessor
from .zerodivision import ZeroDivisionPostProcessor

for arch in LIBVEX_SUPPORTED_ARCHES:
    register(LibVEXLifter, arch)
register(AARCH64Spotter, "AARCH64")
register(ARMSpotter, "ARM")
register(ARMSpotter, "ARMEL")
register(ARMSpotter, "ARMHF")
register(ARMSpotter, "ARMCortexM")
register(AMD64Spotter, "AMD64")
register(X86Spotter, "X86")

__all__ = ["Lifter", "Postprocessor", "lift", "register", "lifters", "ZeroDivisionPostProcessor"]



#!/usr/bin/env python3
import io
import re

STRUCT_R = re.compile(r'^\s+struct {\n$')
FIELD_R = re.compile(r'^\s+unsigned\s+int\s+([a-z\d]*)\s+:\s+(\d+);\n$')
END_R = re.compile(r'\s+}\s+([a-zA-Z\d_]+);\n$')


def work(ss):
    fields = None
    for s in ss:
        if s == '\n':
            continue
        m = STRUCT_R.match(s)
        if m is not None:
            fields = []
            continue
        m = FIELD_R.match(s)
        if m is not None:
            fname, fsize = m.groups()
            fields.append((fname, int(fsize)))
            continue
        m = END_R.match(s)
        if m is not None:
            sname, = m.groups()
            isize = sum(size for _, size in fields)
            if isize == 16 or isize == 32:
                vsize = isize
            elif isize == 48:
                vsize = 64
            else:
                raise Exception(str(isize))
            foffset = 0
            for fname, fsize in fields:
                if len(fname) > 0:
                    voffset = vsize - foffset - fsize
                    vmask = (1 << fsize) - 1
                    insn = '(insn)'
                    if voffset != 0:
                        insn = '(%s >> %d)' % (insn, voffset)
                    insn = '(%s & 0x%x)' % (insn, vmask)
                    print('#define %s_%s(insn) %s' % (sname, fname, insn))
                foffset += fsize
            fields = None
            continue
        raise Exception(s)


work(io.StringIO('''
      struct {
         unsigned int op : 16;
      } E;
      struct {
         unsigned int op :  8;
         unsigned int i  :  8;
      } I;
      struct {
         unsigned int op :  8;
         unsigned int r1 :  4;
         unsigned int r2 :  4;
      } RR;
      struct {
         unsigned int op1 :  8;
         unsigned int r1  :  4;
         unsigned int op2 :  4;
         unsigned int i2  : 16;
      } RI;
      struct {
         unsigned int op : 16;
         unsigned int    :  8;
         unsigned int r1 :  4;
         unsigned int r2 :  4;
      } RRE;
      struct {
         unsigned int op : 16;
         unsigned int r1 :  4;
         unsigned int    :  4;
         unsigned int r3 :  4;
         unsigned int r2 :  4;
      } RRF;
      struct {
         unsigned int op : 16;
         unsigned int m3 :  4;
         unsigned int m4 :  4;
         unsigned int r1 :  4;
         unsigned int r2 :  4;
      } RRF2;
      struct {
         unsigned int op : 16;
         unsigned int r3 :  4;
         unsigned int    :  4;
         unsigned int r1 :  4;
         unsigned int r2 :  4;
      } RRF3;
      struct {
         unsigned int op : 16;
         unsigned int r3 :  4;
         unsigned int    :  4;
         unsigned int r1 :  4;
         unsigned int r2 :  4;
      } RRR;
      struct {
         unsigned int op : 16;
         unsigned int r3 :  4;
         unsigned int m4 :  4;
         unsigned int r1 :  4;
         unsigned int r2 :  4;
      } RRF4;
      struct {
         unsigned int op : 16;
         unsigned int    :  4;
         unsigned int m4 :  4;
         unsigned int r1 :  4;
         unsigned int r2 :  4;
      } RRF5;
      struct {
         unsigned int op :  8;
         unsigned int r1 :  4;
         unsigned int r3 :  4;
         unsigned int b2 :  4;
         unsigned int d2 : 12;
      } RS;
      struct {
         unsigned int op :  8;
         unsigned int r1 :  4;
         unsigned int r3 :  4;
         unsigned int i2 : 16;
      } RSI;
      struct {
         unsigned int op :  8;
         unsigned int r1 :  4;
         unsigned int x2 :  4;
         unsigned int b2 :  4;
         unsigned int d2 : 12;
      } RX;
      struct {
         unsigned int op : 16;
         unsigned int b2 :  4;
         unsigned int d2 : 12;
      } S;
      struct {
         unsigned int op :  8;
         unsigned int i2 :  8;
         unsigned int b1 :  4;
         unsigned int d1 : 12;
      } SI;
      struct {
         unsigned int op1 :  8;
         unsigned int r1  :  4;
         unsigned int r3  :  4;
         unsigned int i2  : 16;
         unsigned int     :  8;
         unsigned int op2 :  8;
      } RIE;
      struct {
         unsigned int op1 :  8;
         unsigned int r1  :  4;
         unsigned int r2  :  4;
         unsigned int i3  :  8;
         unsigned int i4  :  8;
         unsigned int i5  :  8;
         unsigned int op2 :  8;
      } RIE_RRUUU;
      struct {
         unsigned int op1 :  8;
         unsigned int r1  :  4;
         unsigned int     :  4;
         unsigned int i2  : 16;
         unsigned int m3  :  4;
         unsigned int     :  4;
         unsigned int op2 :  8;
      } RIEv1;
      struct {
         unsigned int op1 :  8;
         unsigned int r1  :  4;
         unsigned int r2  :  4;
         unsigned int i4  : 16;
         unsigned int m3  :  4;
         unsigned int     :  4;
         unsigned int op2 :  8;
      } RIE_RRPU;
      struct {
         unsigned int op1 :  8;
         unsigned int r1  :  4;
         unsigned int m3  :  4;
         unsigned int i4  : 16;
         unsigned int i2  :  8;
         unsigned int op2 :  8;
      } RIEv3;
      struct {
         unsigned int op1 :  8;
         unsigned int r1  :  4;
         unsigned int op2 :  4;
         unsigned int i2  : 32;
      } RIL;
      struct {
         unsigned int op1 :  8;
         unsigned int r1  :  4;
         unsigned int m3  :  4;
         unsigned int b4  :  4;
         unsigned int d4  : 12;
         unsigned int i2  :  8;
         unsigned int op2 :  8;
      } RIS;
      struct {
         unsigned int op1 :  8;
         unsigned int r1  :  4;
         unsigned int r2  :  4;
         unsigned int b4  :  4;
         unsigned int d4  : 12;
         unsigned int m3  :  4;
         unsigned int     :  4;
         unsigned int op2 :  8;
      } RRS;
      struct {
         unsigned int op1 :  8;
         unsigned int l1  :  4;
         unsigned int     :  4;
         unsigned int b1  :  4;
         unsigned int d1  : 12;
         unsigned int     :  8;
         unsigned int op2 :  8;
      } RSL;
      struct {
         unsigned int op1 :  8;
         unsigned int r1  :  4;
         unsigned int r3  :  4;
         unsigned int b2  :  4;
         unsigned int dl2 : 12;
         unsigned int dh2 :  8;
         unsigned int op2 :  8;
      } RSY;
      struct {
         unsigned int op1 :  8;
         unsigned int r1  :  4;
         unsigned int x2  :  4;
         unsigned int b2  :  4;
         unsigned int d2  : 12;
         unsigned int m3  :  4;
         unsigned int     :  4;
         unsigned int op2 :  8;
      } RXE;
      struct {
         unsigned int op1 :  8;
         unsigned int r3  :  4;
         unsigned int x2  :  4;
         unsigned int b2  :  4;
         unsigned int d2  : 12;
         unsigned int r1  :  4;
         unsigned int     :  4;
         unsigned int op2 :  8;
      } RXF;
      struct {
         unsigned int op1 :  8;
         unsigned int r1  :  4;
         unsigned int x2  :  4;
         unsigned int b2  :  4;
         unsigned int dl2 : 12;
         unsigned int dh2 :  8;
         unsigned int op2 :  8;
      } RXY;
      struct {
         unsigned int op1 :  8;
         unsigned int i2  :  8;
         unsigned int b1  :  4;
         unsigned int dl1 : 12;
         unsigned int dh1 :  8;
         unsigned int op2 :  8;
      } SIY;
      struct {
         unsigned int op :  8;
         unsigned int l  :  8;
         unsigned int b1 :  4;
         unsigned int d1 : 12;
         unsigned int b2 :  4;
         unsigned int d2 : 12;
      } SS;
      struct {
         unsigned int op :  8;
         unsigned int l1 :  4;
         unsigned int l2 :  4;
         unsigned int b1 :  4;
         unsigned int d1 : 12;
         unsigned int b2 :  4;
         unsigned int d2 : 12;
      } SS_LLRDRD;
      struct {
         unsigned int op :  8;
         unsigned int r1 :  4;
         unsigned int r3 :  4;
         unsigned int b2 :  4;
         unsigned int d2 : 12;
         unsigned int b4 :  4;
         unsigned int d4 : 12;
      } SS_RRRDRD2;
      struct {
         unsigned int op : 16;
         unsigned int b1 :  4;
         unsigned int d1 : 12;
         unsigned int b2 :  4;
         unsigned int d2 : 12;
      } SSE;
      struct {
         unsigned int op1 :  8;
         unsigned int r3  :  4;
         unsigned int op2 :  4;
         unsigned int b1  :  4;
         unsigned int d1  : 12;
         unsigned int b2  :  4;
         unsigned int d2  : 12;
      } SSF;
      struct {
         unsigned int op : 16;
         unsigned int b1 :  4;
         unsigned int d1 : 12;
         unsigned int i2 : 16;
      } SIL;
      struct {
         unsigned int op1 : 8;
         unsigned int v1  : 4;
         unsigned int x2  : 4;
         unsigned int b2  : 4;
         unsigned int d2  : 12;
         unsigned int m3  : 4;
         unsigned int rxb : 4;
         unsigned int op2 : 8;
      } VRX;
      struct {
         unsigned int op1 : 8;
         unsigned int v1  : 4;
         unsigned int v2  : 4;
         unsigned int r3  : 4;
         unsigned int     : 4;
         unsigned int m5  : 4;
         unsigned int     : 4;
         unsigned int m4  : 4;
         unsigned int rxb : 4;
         unsigned int op2 : 8;
      } VRR;
      struct {
         unsigned int op1 : 8;
         unsigned int v1  : 4;
         unsigned int v3  : 4;
         unsigned int i2  : 16;
         unsigned int m3  : 4;
         unsigned int rxb : 4;
         unsigned int op2 : 8;
      } VRI;
      struct {
         unsigned int op1 : 8;
         unsigned int v1  : 4;
         unsigned int v3  : 4;
         unsigned int b2  : 4;
         unsigned int d2  : 12;
         unsigned int m4  : 4;
         unsigned int rxb : 4;
         unsigned int op2 : 8;
      } VRS;
      struct {
         unsigned int op1 : 8;
         unsigned int v1  : 4;
         unsigned int v2  : 4;
         unsigned int b2  : 4;
         unsigned int d2  : 12;
         unsigned int m3  : 4;
         unsigned int rxb : 4;
         unsigned int op2 : 8;
     } VRV;
'''))



import logging
import os
import platform
import re
import subprocess
import sys

import cffi

log = logging.getLogger("cffier")
log.setLevel(logging.DEBUG)


def find_good_scan(questionable):
    known_good = []

    end_line = len(questionable)

    while len(questionable):
        ffi = cffi.FFI()
        log.debug("scan - trying %d good and %d questionable", len(known_good), len(questionable))

        candidate = known_good + questionable[:end_line]
        failed_line = -1

        try:
            ffi.cdef("\n".join(candidate))

            known_good = candidate
            questionable = questionable[end_line:]
            end_line = len(questionable)
        except AssertionError:
            questionable = questionable[1:]
            end_line = len(questionable)
        except cffi.CDefError as e:
            if "<cdef source string>" in str(e):
                failed_line = int(str(e).split("\n")[-1].split(":")[1]) - 1
            elif str(e).count(":") >= 2:
                failed_line = int(str(e).split("\n")[1].split(":")[1])
                failed_line_description = str(e).split("\n")[0]
                idx1 = failed_line_description.index('"')
                idx2 = failed_line_description.rindex('"')
                failed_reason = failed_line_description[idx1 + 1 : idx2]

                for i in range(failed_line, -1, -1):
                    if failed_reason in candidate[i]:
                        failed_line = i
            elif "unrecognized construct" in str(e):
                failed_line = int(str(e).split()[1][:-1]) - 1
            elif "end of input" in str(e):
                end_line -= 1
            else:
                raise Exception("Unknown error")
        except cffi.FFIError as e:
            if str(e).count(":") >= 2:
                failed_line = int(str(e).split("\n")[0].split(":")[1]) - 1
            else:
                raise Exception("Unknown error")

        if failed_line != -1:
            end_line = failed_line - len(known_good)

        if end_line == 0:
            questionable = questionable[1:]
            end_line = len(questionable)
    return known_good


def doit(vex_path):
    cpplist = ["cl", "cpp"]
    cpp = os.getenv("CPP")
    if cpp:
        cpplist.insert(0, cpp)
    if platform.system() == "Darwin":
        cpplist.insert(0, "clang")

    errs = []
    for cpp in cpplist:
        cmd = [cpp, "-I" + vex_path, os.path.join("pyvex_c", "pyvex.h")]
        if cpp in ("cl", "clang", "gcc", "cc", "clang++", "g++"):
            cmd.append("-E")
        try:
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            header, stderr = p.communicate()
            try:
                header = header.decode("utf-8")
                stderr = stderr.decode("utf-8")
            except UnicodeDecodeError:
                # They don't have to be unicode on Windows
                pass

            if not header.strip() or p.returncode != 0:
                errs.append((" ".join(cmd), p.returncode, stderr))
                continue
            else:
                break
        except OSError:
            errs.append((" ".join(cmd), -1, "does not exist"))
            continue
    else:
        log.warning("failed commands:\n" + "\n".join("{} ({}) -- {}".format(*e) for e in errs))
        raise Exception(
            "Couldn't process pyvex headers."
            + 'Please set CPP environmental variable to local path of "cpp".'
            + 'Note that "cpp" and "g++" are different.'
        )
    # header = vex_pp + pyvex_pp

    linesep = "\r\n" if "\r\n" in header else "\n"
    ffi_text = linesep.join(
        line
        for line in header.split(linesep)
        if "#" not in line and line.strip() != "" and "jmp_buf" not in line and not ("=" in line and ";" in line)
    )
    ffi_text = re.sub(r"\{\s*\} NoOp;", "{ int DONOTUSE; } NoOp;", ffi_text)
    ffi_text = re.sub(r"__attribute__\s*\(.*\)", "", ffi_text)
    ffi_text = re.sub(r"__declspec\s*\([^\)]*\)", "", ffi_text)
    ffi_text = ffi_text.replace("__const", "const")
    ffi_text = ffi_text.replace("__inline", "")
    ffi_text = ffi_text.replace("__w64", "")
    ffi_text = ffi_text.replace("__cdecl", "")
    ffi_text = ffi_text.replace("__int64", "long")
    ffi_lines = ffi_text.split(linesep)

    good = find_good_scan(ffi_lines)
    good += ["extern VexControl vex_control;"]

    with open("pyvex/vex_ffi.py", "w") as fp:
        fp.write('ffi_str = """' + "\n".join(good) + '"""\n')
        fp.write("guest_offsets = " + repr(get_guest_offsets(vex_path)) + "\n")


def get_guest_offsets(vex_path):
    fname = os.path.join(vex_path, "libvex_guest_offsets.h")
    out = {}
    with open(fname) as fp:
        for line in fp:
            if line.startswith("#define"):
                _, names, val = line.split()
                val = int(val, 0)
                assert names.startswith("OFFSET_")
                _, arch, reg = names.split("_", 2)
                out[(arch, reg.lower())] = val
    return out


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    doit(sys.argv[1])



from ._register_info import REGISTER_OFFSETS
from .enums import default_vex_archinfo, vex_endness_from_string
from .types import Register
from .vex_ffi import guest_offsets


class PyvexArch:
    """
    An architecture definition for use with pyvex - usable version.
    """

    def __init__(self, name: str, bits: int, memory_endness: str, instruction_endness: str = "Iend_BE"):
        self.name = name
        self.bits = bits
        self.memory_endness = memory_endness
        self.instruction_endness = instruction_endness
        self.byte_width = 8
        self.register_list: list[Register] = []
        self.registers: dict[str, tuple[int, int]] = {}
        self.vex_arch = {
            "X86": "VexArchX86",
            "AMD64": "VexArchAMD64",
            "ARM": "VexArchARM",
            "ARM64": "VexArchARM64",
            "PPC32": "VexArchPPC32",
            "PPC64": "VexArchPPC64",
            "S390X": "VexArchS390X",
            "MIPS32": "VexArchMIPS32",
            "MIPS64": "VexArchMIPS64",
            "RISCV64": "VexArchRISCV64",
        }[name]
        self.ip_offset = guest_offsets[
            (
                self.vex_name_small,
                {
                    "X86": "eip",
                    "AMD64": "rip",
                    "ARM": "r15t",
                    "ARM64": "pc",
                    "PPC32": "cia",
                    "PPC64": "cia",
                    "S390X": "ia",
                    "MIPS32": "pc",
                    "MIPS64": "pc",
                    "RISCV64": "pc",
                }[name],
            )
        ]
        self.vex_archinfo = default_vex_archinfo()
        if memory_endness == "Iend_BE":
            self.vex_archinfo["endness"] = vex_endness_from_string("VexEndnessBE")

    def __repr__(self):
        return f"<PyvexArch {self.name}>"

    @property
    def vex_name_small(self):
        return self.vex_arch[7:].lower()

    def translate_register_name(self, offset, size=None):  # pylint: disable=unused-argument
        for (arch, reg), offset2 in guest_offsets.items():
            if arch == self.vex_name_small and offset2 == offset:
                return reg
        for (arch, reg), offset2 in REGISTER_OFFSETS.items():
            if arch == self.vex_name_small and offset2 == offset:
                return reg
        return str(offset)

    def get_register_offset(self, name: str) -> int:
        arch_reg_tuple = (self.vex_name_small, name)
        if arch_reg_tuple in guest_offsets:
            return guest_offsets[arch_reg_tuple]
        elif arch_reg_tuple in REGISTER_OFFSETS:
            return REGISTER_OFFSETS[arch_reg_tuple]
        else:
            raise KeyError(f"Unknown register {name} for architecture {self.name}")


ARCH_X86 = PyvexArch("X86", 32, "Iend_LE")
ARCH_AMD64 = PyvexArch("AMD64", 64, "Iend_LE")
ARCH_ARM_LE = PyvexArch("ARM", 32, "Iend_LE", instruction_endness="Iend_LE")
ARCH_ARM_BE_LE = PyvexArch("ARM", 32, "Iend_BE", instruction_endness="Iend_LE")
ARCH_ARM_BE = PyvexArch("ARM", 32, "Iend_LE")
ARCH_ARM64_LE = PyvexArch("ARM64", 64, "Iend_LE", instruction_endness="Iend_LE")
ARCH_ARM64_BE = PyvexArch("ARM64", 64, "Iend_BE")
ARCH_PPC32 = PyvexArch("PPC32", 32, "Iend_BE")
ARCH_PPC64_BE = PyvexArch("PPC64", 64, "Iend_BE")
ARCH_PPC64_LE = PyvexArch("PPC64", 64, "Iend_LE")
ARCH_S390X = PyvexArch("S390X", 64, "Iend_BE")
ARCH_MIPS32_BE = PyvexArch("MIPS32", 32, "Iend_BE")
ARCH_MIPS32_LE = PyvexArch("MIPS32", 32, "Iend_LE")
ARCH_MIPS64_BE = PyvexArch("MIPS64", 64, "Iend_BE")
ARCH_MIPS64_LE = PyvexArch("MIPS64", 64, "Iend_LE")
ARCH_RISCV64_LE = PyvexArch("RISCV64", 64, "Iend_LE", instruction_endness="Iend_LE")



from .instr_helper import Instruction
from .lifter_helper import GymratLifter, ParseError
from .syntax_wrapper import VexValue
from .vex_helper import JumpKind, Type

__all__ = [
    "Type",
    "JumpKind",
    "VexValue",
    "ParseError",
    "Instruction",
    "GymratLifter",
    "ParseError",
]



# pylint: disable=no-name-in-module,import-error,missing-class-docstring
import glob
import multiprocessing
import os
import platform
import shutil
import subprocess
import sys
from distutils.command.build import build as st_build
from distutils.util import get_platform

from setuptools import setup
from setuptools.command.develop import develop as st_develop
from setuptools.command.sdist import sdist as st_sdist
from setuptools.errors import LibError

PROJECT_DIR = os.path.dirname(os.path.realpath(__file__))
LIB_DIR = os.path.join(PROJECT_DIR, "pyvex", "lib")
INCLUDE_DIR = os.path.join(PROJECT_DIR, "pyvex", "include")


if sys.platform in ("win32", "cygwin"):
    LIBRARY_FILE = "pyvex.dll"
    STATIC_LIBRARY_FILE = "pyvex.lib"
elif sys.platform == "darwin":
    LIBRARY_FILE = "libpyvex.dylib"
    STATIC_LIBRARY_FILE = "libpyvex.a"
else:
    LIBRARY_FILE = "libpyvex.so"
    STATIC_LIBRARY_FILE = "libpyvex.a"


VEX_LIB_NAME = "vex"  # can also be vex-amd64-linux
VEX_PATH = os.path.abspath(os.path.join(PROJECT_DIR, "vex"))


def _build_vex():
    if len(os.listdir(VEX_PATH)) == 0:
        raise LibError(
            "vex submodule not cloned correctly, aborting.\nThis may be fixed with `git submodule update --init`"
        )

    e = os.environ.copy()
    e["MULTIARCH"] = "1"
    e["DEBUG"] = "1"

    if sys.platform == "win32":
        cmd = ["nmake", "/f", "Makefile-msvc", "all"]
    elif shutil.which("gmake") is not None:
        cmd = ["gmake", "-f", "Makefile-gcc", "-j", str(multiprocessing.cpu_count()), "all"]
    else:
        cmd = ["make", "-f", "Makefile-gcc", "-j", str(multiprocessing.cpu_count()), "all"]

    try:
        subprocess.run(cmd, cwd=VEX_PATH, env=e, check=True)
    except FileNotFoundError as err:
        raise LibError("Couldn't find " + cmd[0] + " in PATH") from err
    except subprocess.CalledProcessError as err:
        raise LibError("Error while building libvex: " + str(err)) from err


def _build_pyvex():
    e = os.environ.copy()
    e["VEX_LIB_PATH"] = VEX_PATH
    e["VEX_INCLUDE_PATH"] = os.path.join(VEX_PATH, "pub")
    e["VEX_LIB_FILE"] = os.path.join(VEX_PATH, "libvex.lib")

    if sys.platform == "win32":
        cmd = ["nmake", "/f", "Makefile-msvc"]
    elif shutil.which("gmake") is not None:
        cmd = ["gmake", "-f", "Makefile", "-j", str(multiprocessing.cpu_count())]
    else:
        cmd = ["make", "-f", "Makefile", "-j", str(multiprocessing.cpu_count())]

    try:
        subprocess.run(cmd, cwd="pyvex_c", env=e, check=True)
    except FileNotFoundError as err:
        raise LibError("Couldn't find " + cmd[0] + " in PATH") from err
    except subprocess.CalledProcessError as err:
        raise LibError("Error while building libpyvex: " + str(err)) from err


def _shuffle_files():
    shutil.rmtree(LIB_DIR, ignore_errors=True)
    shutil.rmtree(INCLUDE_DIR, ignore_errors=True)
    os.mkdir(LIB_DIR)
    os.mkdir(INCLUDE_DIR)

    pyvex_c_dir = os.path.join(PROJECT_DIR, "pyvex_c")

    shutil.copy(os.path.join(pyvex_c_dir, LIBRARY_FILE), LIB_DIR)
    shutil.copy(os.path.join(pyvex_c_dir, STATIC_LIBRARY_FILE), LIB_DIR)
    shutil.copy(os.path.join(pyvex_c_dir, "pyvex.h"), INCLUDE_DIR)
    for f in glob.glob(os.path.join(VEX_PATH, "pub", "*")):
        shutil.copy(f, INCLUDE_DIR)


def _clean_bins():
    shutil.rmtree(LIB_DIR, ignore_errors=True)
    shutil.rmtree(INCLUDE_DIR, ignore_errors=True)


def _build_ffi():
    sys.path.append(".")  # PEP 517 doesn't include . in sys.path
    import make_ffi  # pylint: disable=import-outside-toplevel

    sys.path.pop()

    make_ffi.doit(os.path.join(VEX_PATH, "pub"))


class build(st_build):
    def run(self, *args):
        self.execute(_build_vex, (), msg="Building libVEX")
        self.execute(_build_pyvex, (), msg="Building libpyvex")
        self.execute(_shuffle_files, (), msg="Copying libraries and headers")
        self.execute(_build_ffi, (), msg="Creating CFFI defs file")
        super().run(*args)


class develop(st_develop):
    def run(self):
        self.run_command("build")
        super().run()


class sdist(st_sdist):
    def run(self, *args):
        self.execute(_clean_bins, (), msg="Removing binaries")
        super().run(*args)


cmdclass = {
    "build": build,
    "develop": develop,
    "sdist": sdist,
}

try:
    from setuptools.command.editable_wheel import editable_wheel as st_editable_wheel

    class editable_wheel(st_editable_wheel):
        def run(self):
            self.run_command("build")
            super().run()

    cmdclass["editable_wheel"] = editable_wheel
except ModuleNotFoundError:
    pass

if "bdist_wheel" in sys.argv and "--plat-name" not in sys.argv:
    sys.argv.append("--plat-name")
    name = get_platform()
    if "linux" in name:
        sys.argv.append("manylinux2014_" + platform.machine())
    else:
        # https://www.python.org/dev/peps/pep-0425/
        sys.argv.append(name.replace(".", "_").replace("-", "_"))

setup(cmdclass=cmdclass)



import pyvex


def test_s390x_exrl():
    arch = pyvex.ARCH_S390X
    irsb = pyvex.lift(
        b"\xc6\x10\x00\x00\x00\x04"  # exrl %r1,0x400408
        b"\x07\xfe"  # br %r14
        b"\xd7\x00\x20\x00\x30\x00"  # xc 0(0,%r2),0(%r3)
        b"\x7d\xa7",  # padding
        0x400400,
        arch,
    )
    irsb_str = str(irsb)

    # check last_execute_target, only top 6 bytes are relevant
    assert "0xd700200030000000" in irsb_str
    assert "s390x_dirtyhelper_EX" in irsb_str
    assert "{ PUT(ia) = 0x400400; Ijk_Boring }" in irsb_str
    assert "------ IMark(0x400406, 2, 0) ------" in irsb_str
    assert irsb.jumpkind == "Ijk_Ret"


if __name__ == "__main__":
    test_s390x_exrl()
